from typing import List, Protocol, Set

from judoscale.core.config import Config
from judoscale.core.logger import logger
from judoscale.core.metric import Metric
from judoscale.core.metrics_store import MetricsStore


class Collector(Protocol):
    def collect(self) -> List[Metric]:
        ...

    @property
    def should_collect(self) -> bool:
        ...


class MetricsCollector:
    def __init__(self, config: Config):
        self.config = config

    @property
    def should_collect(self) -> bool:
        return self.config.is_enabled


class WebMetricsCollector(MetricsCollector):
    def __init__(self, config: Config):
        self.store = MetricsStore()
        super().__init__(config=config)

    def add(self, metric: Metric):
        """
        Add metric to the store if it should be collected.
        """
        if self.should_collect:
            self.store.add(metric)

    def collect(self) -> List[Metric]:
        """
        Return all metrics in the store and clear the store.
        """
        return self.store.flush()


class JobMetricsCollector(MetricsCollector):
    """
    Base class for job metrics collectors.
    This class is not meant to be instantiated directly.
    """

    def __init__(self, config: Config):
        super().__init__(config=config)

    @property
    def adapter_config(self):
        raise NotImplementedError("Implement `adapter_config` in a subclass.")

    @property
    def _queues(self) -> List[str]:
        return list()

    @property
    def queues(self) -> Set[str]:
        if configured_queues := self.adapter_config["QUEUES"]:
            return self.limit_max_queues(configured_queues)
        else:
            return self.limit_max_queues(self._queues)

    @property
    def should_collect(self):
        return (
            super().should_collect
            and self.adapter_config["ENABLED"]
            and not self.config["RUNTIME_CONTAINER"].is_redundant_instance
        )

    def limit_max_queues(self, queues: List[str]) -> Set[str]:
        """
        Limit the number of queues to collect metrics for.
        """

        max_queues = self.adapter_config["MAX_QUEUES"]

        if len(queues) > max_queues:
            logger.warning(
                f"{self.__class__.__name__} reporting only {max_queues} queues max, "
                f"skipping the rest ({len(queues) - max_queues})."
            )

        return set(sorted(queues, key=lambda q: len(q))[:max_queues])



from dataclasses import asdict, dataclass
from importlib import metadata
from typing import Optional

from judoscale.core.metrics_collectors import Collector

JUDOSCALE_VERSION = metadata.version("judoscale")


@dataclass
class AdapterInfo:
    """Information about the adapter"""

    platform_version: str
    adapter_version: str = JUDOSCALE_VERSION


@dataclass
class Adapter:
    """Adapter information and optional metrics collector"""

    identifier: str
    adapter_info: AdapterInfo
    metrics_collector: Optional[Collector] = None

    @property
    def as_tuple(self):
        return (self.identifier, asdict(self.adapter_info))



import logging

from django.apps import AppConfig
from django.conf import settings

from judoscale.core.config import config as judoconfig
from judoscale.core.reporter import reporter
from judoscale.django.redis import RedisHelper
from judoscale.rq import judoscale_rq

logger = logging.getLogger(__name__)


class JudoscaleRQConfig(AppConfig, RedisHelper):
    name = "judoscale.rq"
    label = "judoscale_rq"
    verbose_name = "Judoscale (RQ)"

    def ready(self):
        judoconfig.update(getattr(settings, "JUDOSCALE", {}))

        if not judoconfig.is_enabled:
            logger.info("Not activated - No API URL provided")
            return

        judoscale_rq(self.redis_connection(settings.JUDOSCALE.get("REDIS")))
        reporter.ensure_running()



import os
import signal
import sys
import threading
import time
from platform import python_version
from typing import List

import requests

from judoscale.core.adapter import Adapter, AdapterInfo
from judoscale.core.config import Config, config
from judoscale.core.logger import logger
from judoscale.core.metric import Metric
from judoscale.core.metrics_collectors import Collector


class Reporter:
    """
    A reporter that collects metrics from all registered collectors and reports
    them to the Judoscale API.
    """

    def __init__(self, config: Config):
        self.config = config
        self._thread = None
        self._running = False
        self._stopevent = threading.Event()
        self.collectors: List[Collector] = []
        self.adapters: List[Adapter] = []

        self.adapters.append(
            Adapter(
                identifier="judoscale-python",
                adapter_info=AdapterInfo(platform_version=python_version()),
            )
        )

    def add_adapter(self, adapter: Adapter):
        """
        Add an adapter to the reporter.

        If the adapter has a metrics collector, it will be added to the list of
        collectors.
        """
        self.adapters.append(adapter)

        if adapter.metrics_collector:
            self.collectors.append(adapter.metrics_collector)

    def start(self):
        if not self.config.is_enabled:
            logger.info("Reporter not started: API_BASE_URL not set")
            return

        if self.config["RUNTIME_CONTAINER"] is None:
            logger.info("Reporter not started: unknown runtime container")
            return

        if self.config["RUNTIME_CONTAINER"].is_release_instance:
            logger.info("Reporter not started: in a build process")
            return

        logger.info(f"Starting reporter for process {self.pid}")
        self._thread = threading.Thread(target=self._run_loop, daemon=True)
        self._thread.start()
        self._running = True

    def ensure_running(self):
        try:
            if not self.is_running:
                return self.start()
        except Exception as e:
            logger.warning(f"{e.args} - No reporter has initiated")
            pass

    def signal_handler(self, signum, frame):
        self._stopevent.set()
        self._running = False

    @property
    def is_running(self):
        if self._thread and self._thread.is_alive():
            self._running = True
        else:
            self._running = False
        return self._running

    @property
    def pid(self) -> int:
        return os.getpid()

    def _run_loop(self):
        while self.is_running:
            # NOTE: we used to report metrics immediately after the process
            # started, but this caused issues with Celery and Celery Beat
            # when TRACK_BUSY_JOBS was enabled.
            # Instead, we now wait for the first interval to pass before
            # reporting metrics.
            time.sleep(self.config["REPORT_INTERVAL_SECONDS"])
            self._report_metrics()

            if self._stopevent.is_set():
                break

        logger.info(f"{self.pid} reports completed before exiting.")

    @property
    def all_metrics(self) -> List[Metric]:
        """
        Return a list of all metrics collected by all collectors.
        """
        metrics = []
        for collector in self.collectors:
            metrics.extend(collector.collect())
        return metrics

    def _report_metrics(self) -> None:
        report = self._build_report(self.all_metrics)
        url = f"{self.config['API_BASE_URL']}/v3/reports"
        try:
            metrics_length = len(report["metrics"])
            pid = report["pid"]
            logger.debug(
                f"Posting {metrics_length} metrics from {pid} "
                f"to Judoscale adapter API {url}"
            )
            response = requests.post(url, timeout=5, json=report)
            response.raise_for_status()
        except requests.RequestException as e:
            logger.warning("Adapter API request failed - {}".format(e))

    def _build_report(self, metrics: List[Metric]):
        return {
            "container": str(self.config["RUNTIME_CONTAINER"]),
            "pid": self.pid,
            "config": self.config.for_report,
            "adapters": dict(adapter.as_tuple for adapter in self.adapters),
            "metrics": [metric.as_tuple for metric in metrics],
        }


reporter = Reporter(config=config)

previous_handler = signal.getsignal(signal.SIGTERM)


def graceful_shutdown(signum, frame):
    reporter.signal_handler(signum, frame)
    # If there was a previous handler and it's not SIG_DFL or SIG_IGN, call it
    if callable(previous_handler) and previous_handler not in (
        signal.SIG_DFL,
        signal.SIG_IGN,
    ):
        previous_handler(signum, frame)
    else:
        # If no previous handler, use default behavior (exit the process)
        sys.exit(0)


signal.signal(signal.SIGTERM, graceful_shutdown)



from importlib import metadata

from judoscale.core.adapter import Adapter, AdapterInfo
from judoscale.core.config import config as judoconfig
from judoscale.core.metric import Metric
from judoscale.core.metrics_collectors import WebMetricsCollector
from judoscale.core.reporter import reporter


class RequestQueueTimeMiddleware:
    """Django middleware for query metrics report"""

    def __init__(self, get_response):
        self.get_response = get_response
        self.collector = WebMetricsCollector(judoconfig)
        adapter = Adapter(
            identifier="judoscale-django",
            adapter_info=AdapterInfo(platform_version=metadata.version("django")),
            metrics_collector=self.collector,
        )
        reporter.add_adapter(adapter)

    def __call__(self, request):
        request_start_header = request.META.get("HTTP_X_REQUEST_START", "")

        if metric := Metric.for_web(request_start_header):
            self.collector.add(metric)
        reporter.ensure_running()

        return self.get_response(request)



import os

from redis import Redis


class RedisHelper:
    def redis_connection(self, redis_config: dict = None) -> Redis:
        """
        Return a Redis connection from an RQ queue configuration
        """
        if redis_config is not None:
            config = {k.lower(): v for k, v in redis_config.items()}
            if redis_url := config.get("url"):
                del config["url"]
                return Redis.from_url(redis_url, **config)
            else:
                return Redis(**config)
        elif redis_url := os.getenv("REDIS_URL"):
            return Redis.from_url(redis_url)
        else:
            raise RuntimeError(
                "Missing Redis connection configuration. Please set either "
                "settings.JUDOSCALE['REDIS'] or REDIS_URL environment variable."
            )
